"""Symbol constants."""

ASTERISK = "*"
COMMA = ","
DASH = "-"
DOT = "."
EMPTY_STRING = ""
FWD_SLASH = "/"
LOWER_CASE_NA = "na"
NEW_LINE = "\n"
PERIOD = "."
SINGLE_QUOTE = "'"
SPACE = " "
UNDERSCORE = "_"
